import java.applet.Applet;
import java.awt.*;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLEncoder;
import java.net.UnknownHostException;
import java.net.MalformedURLException;
import java.io.IOException;
import java.io.DataInputStream;
import java.util.Vector;


/**
 *
 *  Applet to communicate with the EasyChatServlet.  Implements simple chat 
 *  interface.
 *
 *
 * @version 1.0
 * @author Jim Driscoll jim.driscoll@eng.sun.com
 *
 */
public class EasyChatApplet extends Applet implements Runnable {

    Label userInfo;
    Label messageInfo;
    Label listInfo;
    Button sendButton;
    TextField userText;
    TextArea  messageText;
    List userList;
    URL chatURL;
    URLConnection connect;  
    volatile private boolean loggedin = false;
    String username;
    Thread pollingThread = null;

    /**
     *  Return list of applet parameters (currently null). 
     */
    public String[][] getParameterInfo() {
        return null;
    }

    /**
     *  Return string describing applet.
     */
    public String getAppletInfo() {
            return "EasyChatApplet - Applet to communicate with EasyChatServlet, written by Jim Driscoll";
    }

   
    public synchronized void init() { 
        if (pollingThread != null) return;  //Netscape calls init on any excuse
        super.init();
        resize(500,300);
        userInfo = new Label("Enter User Name (max 10 char):");
        messageInfo = new Label("Message Text:");
        userText = new TextField(40);
        sendButton = new Button("Send");
        messageText = new TextArea(10,40);
        messageText.setEditable(false);

        // Setup main panel and position components with GribBagLayout
        Panel mainp = new Panel();
        GridBagLayout gbl = new GridBagLayout();
        GridBagConstraints gbc = new GridBagConstraints();
        gbc.weightx = 0;
        gbc.weighty = 0;
        gbc.gridx = 0;
        gbc.gridy = 0;
        gbc.gridwidth = 10;
        gbc.gridheight = 1;
        gbc.anchor = GridBagConstraints.CENTER;
        gbc.fill = GridBagConstraints.NONE;
        mainp.setLayout(gbl);
        gbl.setConstraints(userInfo, gbc);
        mainp.add(userInfo);
        gbc.gridy = 1;
        gbc.gridwidth = 9;
        gbc.fill = GridBagConstraints.HORIZONTAL;
        gbl.setConstraints(userText, gbc);
        mainp.add(userText);
        gbc.gridx = 9;
        gbc.gridwidth = 1;
        gbc.fill = GridBagConstraints.NONE;
        gbl.setConstraints(sendButton, gbc);
        mainp.add(sendButton);
        gbc.gridx = 0;
        gbc.gridy = 2;
        gbc.gridwidth = 10;
        gbl.setConstraints(messageInfo, gbc);
        mainp.add(messageInfo);
        gbc.gridy = 3;
        gbc.weighty = 100;
        gbc.gridheight = 10;
        gbc.fill = GridBagConstraints.BOTH;
        gbl.setConstraints(messageText, gbc);
        mainp.add(messageText);

        Panel userp = new Panel();
        userp.setLayout(new BorderLayout());
        listInfo = new Label("Users logged in:");
        userp.add("North",listInfo);
        userList = new List(10, false);
        userList.addItem("Noone logged in"); //15 characters
        userp.add("Center",userList);
        

        //  Add main panel, and userlist to the right.
        setLayout(new BorderLayout());
        add("Center",mainp);
        add("East",userp);

        // Set base URL
        chatURL = getCodeBase();

        // Setup and start polling thread.
        int currPriority = Thread.currentThread().getPriority();
        int newPriority = currPriority == Thread.MIN_PRIORITY ? Thread.MIN_PRIORITY : currPriority - 1;
        pollingThread = new Thread(this,"EasyChatPolling");
        pollingThread.setDaemon(true);
        pollingThread.setPriority(newPriority);
        pollingThread.start();
        p("EasyChat: starting poll");

    }

    /**
     *  Start polling thread if not already running.
     */
    public synchronized void start() {
        if (!isLoggedin() && username != null) {
            login();
        }
        if (pollingThread != null && pollingThread.isAlive()) {
            pollingThread.resume();
            p("EasyChat: resuming poll");
        } else {
            p1("EasyChat: No polling thread!");
            pollingThread = new Thread(this,"EasyChatPolling");
            pollingThread.setDaemon(true);
            pollingThread.start();
            p("EasyChat: starting poll");
        }
    }

    /**
     *  
     */
    public synchronized void stop() {
        if (pollingThread.isAlive()) {
            pollingThread.suspend();
            p("EasyChat: suspending poll");
        } else {
            p1("Easy Chat: Dead polling thread in stop()!");
        }
        logout();
    }

    public synchronized void destroy() {
        if (pollingThread != null && pollingThread.isAlive()) {
            pollingThread.stop();
            pollingThread = null; //just in case we come back
            p("EasyChat: stopping poll");
        }
        logout();
    }

    public void run() {

        p("EasyChat:  starting the polling run");
        while (!Thread.interrupted()) {
            if (isLoggedin()) {
                pollList();
                poll();
                p("EasyChat: polling");
            } else {
                pollList();
                p("EasyChat: not loggedin for poll");
            }
            try {
                Thread.sleep(1000 * 3); // Sleep for 3 seconds
            } catch (InterruptedException e) {}
        }
        p("EasyChat: exiting run()");
    }


    private void login() {
        if (username == null) return;
        String queryString = "chat.html?mode=newuser&user="+URLEncoder.encode(username);
        p("Attempting login as "+username);
        try {
            connect = (new URL(chatURL,queryString)).openConnection();
            connect.setDefaultUseCaches(false); //for future connections
            connect.setUseCaches(false); //for this one
            connect.setDoInput(true);
            connect.setDoOutput(false);
            connect.connect();
            p("Made connection to "+connect);
            DataInputStream in = new DataInputStream(connect.getInputStream());
            String response = in.readLine();
            if (response.startsWith("+")) {
                setLoggedin(true);
                showStatus("Logged into EasyChat as user " + username);
                p("Logged in as "+username);
                userInfo.setText("Type message: ");
                repaint();
            } else {
                showStatus("Error logging in" + response);
                p("Could not log in as "+username);
                System.err.println("Error logging in" +response);
            }
        } catch (MalformedURLException e2) {
            System.err.println("MalformedURLException logging in!");
            e2.printStackTrace(System.err);
            showStatus("Error logging in");
        } catch (IOException e1) {
            System.err.println("IOException logging in!");
            e1.printStackTrace(System.err);
            showStatus("Error logging in");
        }
    }

    private void logout() {
        if (!isLoggedin() || username == null) return;
        String queryString = "chat.html?mode=dropuser&user="+URLEncoder.encode(username);
        try {
            connect = (new URL(chatURL,queryString)).openConnection();
            connect.setUseCaches(false); //for this one
            connect.setDoInput(true);
            connect.setDoOutput(false);
            connect.connect();
            DataInputStream in = new DataInputStream(connect.getInputStream());
            String response = in.readLine();
            if (response.startsWith("+")) {
                setLoggedin(false);
                showStatus("User " + username + " Logged out of EasyChat");
            } else {
                showStatus("Error logging out" + response);
                System.err.println("Error logging out" +response);
            }
        } catch (MalformedURLException e2) {
            System.err.println("MalformedURLException logging in!");
            e2.printStackTrace(System.err);
            showStatus("Error logging out");
        } catch (IOException e1) {
            System.err.println("IOException logging in!");
            e1.printStackTrace(System.err);
            showStatus("Error logging out");
        }
    }

    private void send() {
        String message = userText.getText();
        if (message.equals("")) return;  //don't send an empty message
        userText.setText("");
        showStatus("Sending message");
        String queryString = "chat.html?mode=send&user="+URLEncoder.encode(username);
        queryString = queryString +"&message="+URLEncoder.encode(message);
        try {
            connect = (new URL(chatURL,queryString)).openConnection();
            connect.setUseCaches(false); //for this one
            connect.setDoInput(true);
            connect.setDoOutput(false);
            connect.connect();
            DataInputStream in = new DataInputStream(connect.getInputStream());
            String response = in.readLine();
            if (response.startsWith("+")) {
                showStatus("Message sent");
            } else {
                showStatus("Error sending message" + response);
                System.err.println("Error sending message" + response);
            }
        } catch (MalformedURLException e2) {
            System.err.println("MalformedURLException logging in!");
            e2.printStackTrace(System.err);
            showStatus("Error logging out");
        } catch (IOException e1) {
            System.err.println("IOException logging in!");
            e1.printStackTrace(System.err);
            showStatus("Error logging out");
        }      
    }


    private void poll() {
        String queryString = "chat.html?mode=poll&user="+URLEncoder.encode(username);
        try {
            DataInputStream in = new DataInputStream(new URL(chatURL,queryString).openStream());
            String nextLine = in.readLine();
            if (!nextLine.startsWith("+")) {
                showStatus("Error getting messages from server");
                System.err.println("Error getting messages from server");
                return;
            }
            nextLine = in.readLine();
            while (nextLine != null && !nextLine.equals(".")) {
                System.err.println(nextLine);
                messageText.appendText(nextLine+"\r\n");
                repaint();
                nextLine = in.readLine();
            }
        } catch (IOException e) {
            System.err.println("IOException polling");
            e.printStackTrace(System.err);
            showStatus("Error Checking Messages");
        }

    }

    /**
     *  Poll servlet, asking for list of names, add names to gui list.
     */
    private void pollList() {
        String queryString = "chat.html?mode=list";
        Vector users = new Vector();

        try {
            URL listURL = new URL(chatURL,queryString);
            URLConnection listConn = listURL.openConnection();
            listConn.setDefaultUseCaches(false);
            listConn.setUseCaches(false);
            listConn.connect();
            DataInputStream in = new DataInputStream(listConn.getInputStream());
            String nextLine = in.readLine();
            if (!nextLine.startsWith("+")) {
                showStatus("Error getting userlist from server");
                p1("Error getting userlist from server");
                return;
            }
            nextLine = in.readLine();
            while (nextLine != null && !nextLine.equals(".")) {
                p("Read user: "+nextLine);
                users.addElement(nextLine);
                nextLine = in.readLine();
            }
            if (!users.isEmpty()) {
                userList.clear();
                int size = users.size();
                for (int i = 0; i < size; i++) {
                    userList.addItem((String)users.elementAt(i));
                }
            } else {
                userList.clear();
                userList.addItem("Noone logged in");
            }
            repaint();

        } catch (IOException e) {
            System.err.println("IOException polling");
            e.printStackTrace(System.err);
            showStatus("Error Checking Messages");
        }

    }

    /**
     * Check if user is logged in. 
     */
    public boolean isLoggedin() {
        return loggedin;
    }

    /** 
     * Set loggedin state.
     */
    protected void setLoggedin(boolean newval) {
        loggedin = newval;
    }

    public boolean action(Event evt, Object arg) {
    
        if (evt.target == sendButton || evt.target == userText) {
            if (isLoggedin())
                send();
            else {
                username = userText.getText();
                if (username.length() > 10) {
                    showStatus("10 or fewer characters, please!");
                } else {
                    userText.setText("");
                    login();
                }
            }
            return true;
        }
 
        return super.action(evt,arg);

    }

    private void p(String debug) {
        //System.err.println("EasyChat:"+debug);
    }

    private void p1(String debug) {
        System.err.println("EasyChat:"+debug);
    }

}
