package EasyChat;

import java.servlet.http.HttpServlet;
import java.servlet.http.HttpServletRequest;
import java.servlet.http.HttpServletResponse;
import java.servlet.ServletException;
import java.servlet.ServletConfig;
import java.io.PrintStream;
import java.io.OutputStream;
import java.io.IOException;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;

public class EasyChatServlet extends HttpServlet implements Runnable {

    private UserList users = new UserList();
    private Message message;
    private final static String defaultStartpage = "/EasyChat/EasyChatApplet.html";
    private Thread reaper;
    // Timeout in milliseconds, set to 5 min.
    protected static final int TIMEOUT = 5 * 60 * 1000; 

    public void init(ServletConfig config) throws ServletException {
        super.init(config);

        int currPriority = Thread.currentThread().getPriority();
        int newPriority = currPriority == Thread.MIN_PRIORITY ? Thread.MIN_PRIORITY : currPriority - 1;
        reaper = new Thread(this,"EasyChatReaper");
        reaper.setDaemon(true);
        reaper.setPriority(newPriority);
        reaper.start();

        return;
    }

    public void destroy() {
        super.destroy();
        if (reaper != null && reaper.isAlive()) {
            reaper.stop();
        }
        return;
    }

    public void service(HttpServletRequest req, HttpServletResponse res) 
      throws IOException, ServletException {

        String user, mode, usermsg;

        if (req.getQueryString() == null) {  //return starting page...
            if (getInitParameter("startpage") == null)
                res.sendRedirect(defaultStartpage);
            res.sendRedirect(getInitParameter("startpage"));
            return;
        }
        
        mode = req.getParameter("mode");
        user = req.getParameter("user");
        if (mode == null) {
            sendError(res,"Mode not set");
            return;
        }
        if (user == null && !mode.equals("list")) {
            sendError(res,"User not set");
            return;
        }
        if (mode.equals("newuser")) {
            if (users.exists(user)) {
                sendError(res,"User exists");
                return;
            }
            if (!users.add(user)) {
                sendError(res,"Problem adding user");
                return;
            }
            users.addMessage(new Message("EasyChat",user + " has joined the chat."));
            log("The user "+user+" logged in.");
            sendResp(res,"User logged in.");
            return;
        } else if (mode.equals("dropuser")) {
            if (!users.exists(user)) {
                sendError(res,"User does not exist");
                return;
            }
            if (!users.drop(user)) {
                sendError(res,"Problem deleting user");
                return;
            }
            users.addMessage(new Message("EasyChat",user + " has left the chat."));       
            log("The user "+user+" logged out.");
            sendResp(res,"User logged out.");
            return;
        } else if (mode.equals("poll")) {
            if (!users.exists(user)) {
                sendError(res,"Invalid user");
                return;
            }
            returnMessages(res,users.getMessages(user));
            users.resetUser(user);
            return;
        } else if (mode.equals("send")) {
            usermsg = req.getParameter("message");
            if (usermsg == null) {
                sendError(res,"Message not set for send");
                return;
            }
            users.addMessage(new Message(user,usermsg));
            sendResp(res,"Message Added.");
            return;
        } else if (mode.equals("list")) {
            returnUserList(res);
            return;
        } else {
            sendError(res,"Invalid Mode.");
            return;
        }
    }
    
    public void run() {
        while (!Thread.interrupted()) {
            Enumeration inactive = users.inactive(TIMEOUT);
            while (inactive.hasMoreElements()) {
                String user = (String)inactive.nextElement();
                users.drop(user);
                users.addMessage(new Message("EasyChat",user + " has timed out."));
                users.addMessage(new Message("EasyChat","Dropping "+user+" from the chat."));
                log("User "+user+" timed out.  Dropped from chat.");
            }
            try {
                Thread.sleep(TIMEOUT); // Sleep for TIMEOUT
            } catch (InterruptedException e) {}
        }
    }

    protected void sendError(HttpServletResponse res, String error) {
        try {
            PrintStream pout = new PrintStream(res.getOutputStream());
            pout.print("-ERR "+error+"\r\n");
            pout.flush();
            pout.close();
        } catch (IOException e) {
            // Silently ignore failure to send.  For now.
        }
    }

    protected void sendResp(HttpServletResponse res, String resp) {
        try {
            res.setDateHeader("Expires",System.currentTimeMillis());
            PrintStream pout = new PrintStream(res.getOutputStream());
            pout.print("+OK "+resp+"\r\n");
            pout.flush();
            pout.close();
        } catch (IOException e) {
            // Silently ignore failure to send.  For now.
        }
    }

    protected void returnMessages(HttpServletResponse res, Enumeration messagelist) {
        try {
            res.setDateHeader("Expires",System.currentTimeMillis());
            PrintStream pout = new PrintStream(res.getOutputStream());
            pout.print("+OK\r\n");
            while (messagelist.hasMoreElements()) {
                Message message = (Message) messagelist.nextElement();
                pout.print(message+"\r\n");
            }
            pout.print(".\r\n");
            pout.flush();
            pout.close();
        } catch (IOException e) {
            // Silently ignore failure to send.  For now.
        }
    }

    protected void returnUserList(HttpServletResponse res) {
      try {
          res.setDateHeader("Expires",System.currentTimeMillis());
          PrintStream pout = new PrintStream(res.getOutputStream());
          pout.print("+OK\r\n");
          Enumeration userlist = users.list();
          while (userlist.hasMoreElements()) {
              pout.print((String)userlist.nextElement());
              pout.print("\r\n");
          }
          pout.print(".\r\n");
          pout.flush();
      } catch (IOException e) {
          //Silently ignore failure to send.  For now.
      }
    }
}


/**
 *  Encapsulate a list of users, including time last active.
 *
 *  @author Jim Driscoll
 *  @version 1.2 3/7/97
 */
class UserList {

    private Hashtable list = new Hashtable();
    private Hashtable active = new Hashtable();

    UserList() {
    }

    protected synchronized boolean add(String user) {
        if (exists(user)) return false;
        list.put(user, new MessageQueue());
        touch(user);
        return true;
    }

    protected synchronized Enumeration list() {
        return list.keys();
    }

    protected synchronized boolean drop(String user) {
        //should I even check?
        if (!exists(user)) return false;
        list.remove(user);
        active.remove(user);
        return true;
    }

    /**
     * Check the existance of a user.
     */
    protected boolean exists(String user) {
        return list.containsKey(user);
    }

    protected void addMessage(Message message) {
        Enumeration allUsers = list.keys();
        while (allUsers.hasMoreElements()) 
            addUserMessage((String) allUsers.nextElement(),message);
    }

    private void addUserMessage(String user, Message message) {

        ((MessageQueue)list.get(user)).add(message);        
    }

    protected void resetUser(String user) {
        ((MessageQueue)list.get(user)).reset();
    }

    /**
     * Return an Enumeration of messages waiting for the user.
     */
    protected Enumeration getMessages(String user) {
        touch(user);
        return ((MessageQueue)list.get(user)).listMessages();
    }

    /**
     * Update last active time for a user.
     */
    protected synchronized void touch(String user) {
        if (exists(user)) {
            active.put(user, new Long(System.currentTimeMillis()));
        }
    }

    /** 
     *  Return an Enumeration of inactive users, as specified by timeout in
     *  milliseconds.  This is a pretty heavyweight process for this class.
     *  @return time millisecond timeout value
     */
    protected synchronized Enumeration inactive(int timeout) {
        Vector inactive = new Vector();

        long now = System.currentTimeMillis();
        long when = now - timeout;

        Enumeration users = list.keys();
        while (users.hasMoreElements()) {
            String user = (String)users.nextElement();
            if (((Long)active.get(user)).longValue() < when) {
                inactive.addElement(user);
            }
        }
        return inactive.elements();
    }
}

/** 
 * 
 * A list of messages. 
 *
 */
class MessageQueue {

    private Vector messageq = new Vector();

    MessageQueue() {}

    /**
     *  Add a message to the queue.
     */
    protected synchronized void add(Message message) {
        messageq.addElement(message);
        notifyAll();
    }

    /**
     *  Retrieve a list of messages in the queue. Blocks until a
     *  message is waiting to be sent.
     */
    protected synchronized Enumeration listMessages() {
        try {
            if (messageq.isEmpty()) {
                wait(30*1000); //30 second timeout
            }
        } catch (InterruptedException ie) {
            //If interrupted, fall through
        }
        return messageq.elements();
    }

    protected synchronized void reset() {
         messageq.removeAllElements(); 
    }
}
/**
 *  Message.  What else?  Associates user and what they said.
 *  Note:  consider adding access classes and timestamp
 *
 */
class Message {
    private String user;
    private String message;

    Message(String user, String message) {
        this.user = user;
        this.message = message;
    }

    public String toString() {
        return user+": "+message;
    }

    protected String getUser() {
        return user;
    }

    protected String getMessage() {
        return message;
    }
}
